(*. (c) by Richard Lang, 2003 .*)
(* theory collecting all knowledge for LinearEquations
   created by: rlang 
         date: 02.10
   changed by: rlang
   last change by: rlang
             date: 02.10.20
*)

theory LinEq imports Poly Equation begin

section \<open>Theorems for Rule_Set; FIXME: replace axiomatization\<close>

axiomatization where
(*-- normalise --*)
  (*WN0509 compare PolyEq.all_left "[|Not(b=!=0)|] ==> (a = b) = (a - b = 0)"*)
  all_left:          "[|Not(b=!=0)|] ==> (a=b) = (a+(-1)*b=0)" and
  makex1_x:          "a \<up> 1  = a" and
  real_assoc_1:      "a+(b+c) = a+b+c" and
  real_assoc_2:      "a*(b*c) = a*b*c" and

(*-- solve --*)
  lin_isolate_add1:  "(a + b*bdv = 0) = (b*bdv = (-1)*a)" and
  lin_isolate_add2:  "(a +   bdv = 0) = (  bdv = (-1)*a)" and
  lin_isolate_div:   "[|Not(b=0)|] ==> (b*bdv = c) = (bdv = c / b)"


section \<open>Rule_Set for Problem and MethodC\<close>

ML \<open>
val LinEq_prls = (*3.10.02:just the following order due to subterm evaluation*)
  Rule_Set.append_rules "LinEq_prls" Rule_Set.empty [
    \<^rule_eval>\<open>HOL.eq\<close> (Prog_Expr.eval_equal "#equal_"),
    \<^rule_eval>\<open>Prog_Expr.matches\<close> (Prog_Expr.eval_matches ""),
    \<^rule_eval>\<open>Prog_Expr.lhs\<close> (Prog_Expr.eval_lhs ""),
    \<^rule_eval>\<open>Prog_Expr.rhs\<close> (Prog_Expr.eval_rhs ""),
    \<^rule_eval>\<open>has_degree_in\<close> (eval_has_degree_in ""),
    \<^rule_eval>\<open>is_polyrat_in\<close> (eval_is_polyrat_in ""),
    \<^rule_eval>\<open>Prog_Expr.occurs_in\<close> (Prog_Expr.eval_occurs_in ""),    
    \<^rule_eval>\<open>Prog_Expr.ident\<close> (Prog_Expr.eval_ident "#ident_"),
    \<^rule_thm>\<open>not_true\<close>,
    \<^rule_thm>\<open>not_false\<close>,
    \<^rule_thm>\<open>and_true\<close>,
    \<^rule_thm>\<open>and_false\<close>,
    \<^rule_thm>\<open>or_true\<close>,
    \<^rule_thm>\<open>or_false\<close>];

(* ----- asm_rls ----- *)
val LinEq_crls = 
   Rule_Set.append_rules "LinEq_crls" poly_crls
   [\<^rule_thm>\<open>real_assoc_1\<close>
    (*		
     Don't use
     \<^rule_eval>\<open>divide\<close> (Prog_Expr.eval_cancel "#divide_e"),
     \<^rule_eval>\<open>realpow\<close> (**)(Calc_Binop.numeric "#power_"),
     *)
    ];

(* ----- crls ----- *)
val LinEq_erls = 
  Rule_Set.append_rules "LinEq_erls" Poly_erls [
    \<^rule_thm>\<open>real_assoc_1\<close>
    (*		
     Don't use
     \<^rule_eval>\<open>divide\<close> (Prog_Expr.eval_cancel "#divide_e"),
     \<^rule_eval>\<open>realpow\<close> (**)(Calc_Binop.numeric "#power_"),
     *)
    ];
\<close>
rule_set_knowledge LinEq_erls = LinEq_erls
ML \<open>
    
val LinPoly_simplify = prep_rls'(
  Rule_Def.Repeat {id = "LinPoly_simplify", preconds = [], 
    rew_ord = ("termlessI",termlessI), 
    asm_rls = LinEq_erls, 
    prog_rls = Rule_Set.Empty, 
    calc = [], errpatts = [],
    rules = [
      \<^rule_thm>\<open>real_assoc_1\<close>,
      \<^rule_eval>\<open>plus\<close> (**)(Calc_Binop.numeric "#add_"),
      \<^rule_eval>\<open>minus\<close> (**)(Calc_Binop.numeric "#sub_"),
      \<^rule_eval>\<open>times\<close> (**)(Calc_Binop.numeric "#mult_"),
      (*  Don't use  
       \<^rule_eval>\<open>divide\<close> (Prog_Expr.eval_cancel "#divide_e"),
       \<^rule_eval>\<open>sqrt\<close> (eval_sqrt "#sqrt_"),
       *)
      \<^rule_eval>\<open>realpow\<close> (**)(Calc_Binop.numeric "#power_")],
    program = Rule.Empty_Prog});
\<close>
rule_set_knowledge LinPoly_simplify = LinPoly_simplify

ML \<open>
(*isolate the bound variable in an linear equation; 'bdv' is a meta-constant*)
val LinEq_simplify = prep_rls'(
  Rule_Def.Repeat {id = "LinEq_simplify", preconds = [],
    rew_ord = ("Rewrite_Ord.id_empty", Rewrite_Ord.function_empty),
    asm_rls = LinEq_erls,
    prog_rls = Rule_Set.Empty,
    calc = [], errpatts = [],
    rules = [
	     \<^rule_thm>\<open>lin_isolate_add1\<close>, (* a+bx=0 -> bx=-a *)
	     \<^rule_thm>\<open>lin_isolate_add2\<close>, (* a+ x=0 ->  x=-a *)
	     \<^rule_thm>\<open>lin_isolate_div\<close> (*   bx=c -> x=c/b *)],
    program = Rule.Empty_Prog});
\<close>
rule_set_knowledge LinEq_simplify = LinEq_simplify


section \<open>Problem.T\<close>

problem pbl_equ_univ_lin : "LINEAR/univariate/equation" =
  \<open>LinEq_prls\<close>
  Method_Ref: "LinEq/solve_lineq_equation"
  CAS: "solve (e_e::bool, v_v)"
  Given: "equality e_e" "solveFor v_v"
  Where: "HOL.False" (*WN0509 just detected: this pbl can never be used?!?*)
    "Not( (lhs e_e) is_polyrat_in v_v)"
    "Not( (rhs e_e) is_polyrat_in v_v)"
    "((lhs e_e) has_degree_in v_v)=1"
	  "((rhs e_e) has_degree_in v_v)=1"
  Find: "solutions v_v'i'"


section \<open>MethodC.T\<close>

method met_eqlin : "LinEq" =
  \<open>{rew_ord = "tless_true",rls' = Atools_erls,calc = [], prog_rls = Rule_Set.empty, where_rls = Rule_Set.empty,
    errpats = [], rew_rls = norm_Poly}\<close>
    (* ansprechen mit ["LinEq", "solve_univar_equation"] *)

partial_function (tailrec) solve_linear_equation :: "bool \<Rightarrow> real \<Rightarrow> bool list"
  where
"solve_linear_equation e_e v_v = (
  let
    e_e = (
      (Try (Rewrite ''all_left'')) #>
      (Try (Repeat (Rewrite ''makex1_x''))) #>
      (Try (Rewrite_Set ''expand_binoms'')) #>
      (Try (Repeat (Rewrite_Set_Inst [(''bdv'', v_v)] ''make_ratpoly_in''))) #>
      (Try (Repeat (Rewrite_Set ''LinPoly_simplify''))) ) e_e;
    e_e = (
      (Try (Rewrite_Set_Inst [(''bdv'', v_v)] ''LinEq_simplify'')) #>
      (Repeat (Try (Rewrite_Set ''LinPoly_simplify''))) ) e_e
  in
    Or_to_List e_e)"

method met_eq_lin : "LinEq/solve_lineq_equation" =
  \<open>{rew_ord = "termlessI", rls' = LinEq_erls, prog_rls = Rule_Set.empty, where_rls = LinEq_prls, calc = [],
    errpats = [], rew_rls = norm_Poly}\<close>
  Program: solve_linear_equation.simps
  Given: "equality e_e" "solveFor v_v"
  Where: "Not ((lhs e_e) is_polyrat_in v_v)" "((lhs e_e) has_degree_in v_v) = 1"
  Find: "solutions v_v'i'"

section \<open>For trials\<close>
ML \<open>
\<close> ML \<open>
\<close> ML \<open>
\<close>

end

