(* Title:  solve an example by interpreting a method's script
   Author: Walther Neuper 1999
   (c) copyright due to lincense terms.
*)

signature SOLVE_PHASE =
sig
  datatype auto = CompleteCalc | CompleteCalcHead | CompleteModel | CompleteSubpbl 
  | CompleteToSubpbl | Steps of int
  val autoord : auto -> int

  val all_solve : auto -> Pos.pos' list -> Calc.T ->
    string * Pos.pos' list * (Ctree.ctree * (int list * Pos.pos_))
  val complete_solve :
     auto -> Pos.pos' list -> Calc.T -> string * Pos.pos' list * Calc.T
end

(**)
structure Solve(**): SOLVE_PHASE(**) =
struct
(**)
open Ctree;
open Pos

type squ = ctree; (* TODO: safe etc. *)

(* tell how may steps of a calculation should be done by "fun autocalc"
   FIXXXME040624: does NOT match interfaces/ITOCalc.java
   TODO.WN0512 redesign togehter with autocalc ?                     *)
datatype auto = 
  Steps of int      (*1 do #int steps (may stop in model/specify)
                       IS VERY INEFFICIENT IN MODEL/SPECIY                    *)
| CompleteModel    (*2 complete modeling
                       if model complete, finish specifying                   *)
| CompleteCalcHead (*3 complete model/specify in one go                       *)
| CompleteToSubpbl (*4 stop at the next begin of a subproblem,
                       if none, complete the actual (sub)problem              *)
| CompleteSubpbl   (*5 complete the actual (sub)problem (incl.ev.subproblems) *)
| CompleteCalc;    (*6 complete the calculation as a whole                    *)

fun autoord (Steps _ ) = 1
  | autoord CompleteModel = 2
  | autoord CompleteCalcHead = 3
  | autoord CompleteToSubpbl = 4
  | autoord CompleteSubpbl = 5
  | autoord CompleteCalc = 6;

fun complete_solve auto c (ptp as (_, p as (_, p_))) =
    if p = ([], Res)
    then ("end-of-calculation", [], ptp)
    else
      if member op = [Pbl, Met] p_
      then
        let
          val ptp = Specify.do_all ptp
	        val (_, c', ptp) = all_solve auto c ptp
	      in complete_solve auto (c @ c') ptp end
      else
        case Step_Solve.do_next ptp of
	        (_, ((Tactic.Subproblem _, _, _) :: _, c', ptp')) =>
	          if autoord auto < 5
            then ("ok", c @ c', ptp)
	          else
              let
                val ptp = Specify.do_all ptp'
	              val (_, c'', ptp) = all_solve auto (c @ c') ptp
	            in complete_solve auto (c @ c'@ c'') ptp end
	      | (_, ((Tactic.Check_Postcond _, _, _) :: _, c', ptp' as (_, p'))) =>
	        if autoord auto < 6 orelse p' = ([], Res)
          then ("ok", c @ c', ptp')
	        else complete_solve auto (c @ c') ptp'
	      | (_, ((Tactic.End_Detail, _, _) :: _, c', ptp')) => 
	        if autoord auto < 6
          then ("ok", c @ c', ptp')
	        else complete_solve auto (c @ c') ptp'
	      | (_, (_, c', ptp')(*fall through*)) => complete_solve auto (c @ c') ptp'
and all_solve auto c (ptp as (pt, pos as (p,_)): ctree * pos') = 
    let
      val (_, _, mI) = get_obj g_spec pt p
      val ctxt = get_ctxt pt pos
      val (_, (_, c', ptp)) = LI.by_tactic (Tactic.Apply_Method' (mI, NONE, Istate.empty, ctxt)) (Istate.empty, ctxt) ptp
    in
      complete_solve auto (c @ c') ptp
    end;

end